use crate::core::input::InputManager;
use crate::core::render::primitives::camera::Camera;
use crate::core::render::render_context::RenderContext;
use crate::core::render::texture::Texture;
use crate::core::scene::Scene;
use crate::core::timer::Timer;
use glam::{Mat4, Vec3};
use vulkano::command_buffer::{
    AutoCommandBufferBuilder, CommandBufferUsage, PrimaryAutoCommandBuffer,
    PrimaryCommandBufferAbstract,
};

use super::assets::square::Square;

pub struct MainSceneState {
    square: Square,
    camera: Camera,
    texture: Texture,
    speed: f32,
}

#[derive(Default)]
pub struct MainScene {
    state: Option<MainSceneState>,
}

impl Scene for MainScene {
    fn loaded(&self) -> bool {
        self.state.is_some()
    }

    fn load(&mut self, render_context: &RenderContext) {
        let square = Square::new(
            render_context.device(),
            render_context.memory_allocator(),
            render_context.swapchain_format(),
        )
        .unwrap();

        let camera = Camera::new(
            Mat4::look_at_rh(
                Vec3::new(0.3, 0.3, 1.0),
                Vec3::new(0.0, 0.0, 0.0),
                Vec3::new(0.0, -1.0, 0.0),
            ),
            Mat4::perspective_rh_gl(
                std::f32::consts::FRAC_PI_2,
                render_context.aspect_ratio(),
                0.01,
                100.0,
            ),
        );

        let mut uploads = AutoCommandBufferBuilder::primary(
            render_context.command_buffer_allocator().clone(),
            render_context.graphics_queue().queue_family_index(),
            CommandBufferUsage::OneTimeSubmit,
        )
        .unwrap();

        let texture = Texture::from_file(
            render_context.device(),
            render_context.memory_allocator(),
            &mut uploads,
            "res/textures/wooden-crate.jpg",
        )
        .unwrap();

        let _ = uploads
            .build()
            .unwrap()
            .execute(render_context.graphics_queue().clone())
            .unwrap();

        self.state = Some(MainSceneState {
            square,
            camera,
            texture,
            speed: 50.0,
        });
    }

    fn update(
        &mut self,
        render_context: &RenderContext,
        input_manager: &InputManager,
        timer: &Timer,
    ) {
        let state = self.state.as_mut().unwrap();
        state.camera.update(input_manager, timer, state.speed, 10.0);

        state.camera.set_projection(Mat4::perspective_rh_gl(
            std::f32::consts::FRAC_PI_2,
            render_context.aspect_ratio(),
            0.01,
            100.0,
        ));
    }

    fn render(
        &self,
        render_context: &RenderContext,
        builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>,
    ) {
        let state = self.state.as_ref().unwrap();
        let camera_uniform = state
            .camera
            .create_buffer(render_context.memory_allocator())
            .unwrap();

        state
            .square
            .render(
                builder,
                render_context.descriptor_set_allocator(),
                &camera_uniform,
                &state.texture,
            )
            .unwrap();
    }

    fn unload(&mut self) {}
}
